from abc import ABCMeta
from abc import abstractmethod
import logging

from sklearn.base import BaseEstimator
from skorch import NeuralNet

logger = logging.getLogger(__name__)


def filter_dict_by_prefix(source, prefix):
    result = dict()
    for key in source.keys():
        if key.startswith(prefix):
            key_stripped = key[len(prefix) :]
            result[key_stripped] = source[key]
    return result


class Learner(BaseEstimator, metaclass=ABCMeta):
    def _store_kwargs(self, kwargs, allowed_prefixes):
        """Store kwargs with whitelisted prefixes into this objects attributes.

        Raises an exception if one of the kwargs does not match a whiltelisted prefix.
        """

        def starts_with_legal_prefix(key):
            for prefix in allowed_prefixes:
                if key.startswith(prefix):
                    return True
            return False

        for key in kwargs.keys():
            if not starts_with_legal_prefix(key):
                raise TypeError(
                    f"'__init__() got an unexpected keyword argument '{key}'. Allowed prefixes: {allowed_prefixes}."
                )

        vars(self).update(kwargs)

    def _get_prefix_attributes(self, prefix):
        """Return all attributes of this class that start with a given prefix.

        The prefix is stripped in the result. This can be used to pass on some
        parameters to subclasses.
        """
        return filter_dict_by_prefix(self.__dict__, prefix)

    @abstractmethod
    def fit(self, X, Y, **kwargs):
        """
        Fit the preference learning algorithm on the provided set of queries X and preferences Y of those objects.
        The provided queries and corresponding preferences are of a fixed size (numpy arrays).

        Parameters
        ----------
        X : array-like, shape (n_samples, n_objects, n_features)
            Feature vectors of the objects
        Y : array-like, shape (n_samples, n_objects)
            Preferences of the objects in form of rankings or choices
        """
        raise NotImplementedError

    def _pre_fit(self):
        """Perform stateful initialization before fitting.

        This function is for initialization that does not depend on the data,
        but still requires some processing and therefore should not happen in
        __init__. Examples include initialization of optimizers, construction
        of NeuralNetwork layers (if it can be done without knowledge of the
        data) etc.

        You should always call this function before fit, even if you do not
        override it. If you override it, you should call the super method first
        so that general initializations can be inherited.
        """
        pass

    @abstractmethod
    def _predict_scores_fixed(self, X, **kwargs):
        """
        Predict the scores for a given collection of sets of objects of same size.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_objects, n_features)


        Returns
        -------
        Y : array-like, shape (n_samples, n_objects)
            Returns the scores of each of the objects for each of the samples.
        """
        raise NotImplementedError

    def predict_for_scores(self, scores, **kwargs):
        raise NotImplementedError

    def predict_scores(self, X, **kwargs):
        """
        Predict the utility scores for each object in the collection of set of objects called a query set.

        Parameters
        ----------
        X : dict or numpy array
            Dictionary with a mapping from query set size to numpy arrays or a single numpy array of size:
            (n_instances, n_objects, n_features)

        Returns
        -------
        Y : dict or numpy array
            Dictionary with a mapping from query set size to numpy arrays or a single numpy array of size:
            (n_instances, n_objects)
        """
        logger.info("Predicting scores")

        if isinstance(X, dict):
            scores = dict()
            for ranking_size, x in X.items():
                n_instances, n_objects, n_features = x.shape
                scores[ranking_size] = self._predict_scores_fixed(x, **kwargs)

        else:
            scores = self._predict_scores_fixed(X, **kwargs)
        return scores

    def predict(self, X, **kwargs):
        """
        Predict preferences in the form of rankings or choices for a given collection of sets of objects called
        a query set using the function :meth:`.predict_for_scores`.

        Parameters
        ----------
        X : dict or numpy array
            Dictionary with a mapping from the query set size to numpy arrays or a single numpy array of size:
            (n_instances, n_objects, n_features)


        Returns
        -------
        Y : dict or numpy array
            Dictionary with a mapping from the query set size to numpy arrays or a single numpy array containing
            predicted preferences of size:
            (n_instances, n_objects)
        """
        logger.debug("Predicting started")

        scores = self.predict_scores(X, **kwargs)
        logger.debug("Predicting scores complete")

        return self.predict_for_scores(scores, **kwargs)

    def __call__(self, X, *args, **kwargs):
        """
        Predict preferences in the form of rankings or choices for a given collection of sets of objects.

        Parameters
        ----------
        X : dict or numpy array
            Dictionary with a mapping from ranking size to numpy arrays
            or a single numpy array of size:
            (n_instances, n_objects, n_features)


        Returns
        -------
        Y : Dictionary with a mapping from ranking size to numpy arrays
            or a single numpy array of size:
            (n_instances, n_objects)
            Predicted ranking
        """
        return self.predict(X, **kwargs)

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Learner:
            has_fit = any("fit" in B.__dict__ for B in C.__mro__)
            has_predict = any("predict" in B.__dict__ for B in C.__mro__)
            has_scores = any("predict_scores" in B.__dict__ for B in C.__mro__)
            has_scores_fixed = any(
                "_predict_scores_fixed" in B.__dict__ for B in C.__mro__
            )
            has_predict_for_scores = any(
                "predict_for_scores" in B.__dict__ for B in C.__mro__
            )
            if (
                has_fit
                and has_predict
                and has_scores
                and has_scores_fixed
                and has_predict_for_scores
            ):
                return True
        return NotImplemented


class SkorchInstanceEstimator(NeuralNet, Learner):
    """Base estimator for torch-based ranking and choice tasks.

    This establishes the basic interface of a cs-ranking learner that is
    compatible with scikit-learn. It is based on an skorch estimator with the
    added assumption that the ``module`` expects the number of features per
    object as a parameter. The ``module`` should then predict a score for each
    object which can later be converted to a prediction (i.e. a ranking, a
    general choice or a discrete choice). To derive a new estimator you should
    therefore override the constructor to set default values for the ``module``
    and the ``criterion`` parameter. You should also override the
    ``predict_for_scores`` function to specify how the scores can be converted
    to the target prediction. You may use one of the existing mixins such as
    ``ObjectRanker`` for that purpose.

    See the documentation of ``skorch.NeuralNet`` for a description of the
    possible parameters.
    """

    def _get_extra_module_parameters(self):
        """Return extra parameters that should be passed to the module.

        You should take care to update the dictionary from the ``super``
        implementation when overriding this function. You usually do not want
        to just discard the parameters that are specified by the super class.
        """
        return {"n_features": self.n_features_}

    def get_params_for(self, prefix):
        """Return the init parameters for an attribute.

        This extends the ``get_params_for`` function from skorch to inject
        custom module parameters. This allows us to pass parameters that do not
        directly correspond to parameters of this estimator while also sticking
        to the scikit-learn estimator API. Overriding this function is
        preferable to than overriding ``initialize_module`` since this function
        does not modify the object's state and we can simply extend the results
        of a ``super`` delegation.
        """
        params = super().get_params_for(prefix)

        if prefix == "module":
            # Explicitly set parameters override the default values.
            defaults = self._get_extra_module_parameters()
            defaults.update(params)
            return defaults
        else:
            return params
        return params

    def fit(self, X, y=None, **fit_params):
        """Fit the estimator to data.

        This derives the number of object features from the data and then
        delegates to ``skorch.NeuralNet.fit``. See the documentation of that
        method for more details.

        Parameters
        ----------
        X : input data
            May take various forms, such as numpy arrays or torch datasets. See
            the documentation of ``skorch.NeuralNet.fit`` for more details.

        y : target data
            May take the same forms as ``x``. This is optional since the target
            data may already be included in the data structure that is passed
            as ``X``. See the documentation of ``skorch.NeuralNet.fit`` for
            more details.

        **fit_params : dict
            Additional fit parameters. See the documentation of
            ``skorch.NeuralNet.fit`` for more details.
        """
        dataset = self.get_dataset(X, y)
        (_n_objects, self.n_features_) = dataset[0][0].shape
        NeuralNet.fit(self, X=dataset, y=None, **fit_params)

    def predict(self, X, **kwargs):
        """Predict targets for inputs.

        This delegates to ``csrank.Learner.predict``. See the documentation of
        that function for details.

        Parameters
        ----------
        X : dict or numpy array
            Dictionary with a mapping from the query set size to numpy arrays or a single numpy array of size:
            (n_instances, n_objects, n_features)

        Returns
        -------
        Y : dict or numpy array
            Dictionary with a mapping from the query set size to numpy arrays or a single numpy array containing
            predicted preferences of size:
            (n_instances, n_objects)
        """
        return Learner.predict(self, X, **kwargs)

    def _predict_scores_fixed(self, X, **kwargs):
        """Predict scores for a collection of sets of objects of the same size.

        This simply queries the torch module for a prediction on the input
        data, which can then be interpreted as scores.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_objects, n_features)
            The input data.

        Returns
        -------
        Y : array-like, shape (n_samples, n_objects)
            The predicted scores.
        """
        return self.predict_proba(X, **kwargs)



import numpy as np
from pymc3.variational.callbacks import CheckParametersConvergence
import pytest
import torch
from torch import optim

from csrank.constants import CMPNET_DC
from csrank.constants import FATE_DC
from csrank.constants import FETA_DC
from csrank.constants import GEV
from csrank.constants import MLM
from csrank.constants import MNL
from csrank.constants import NLM
from csrank.constants import PCL
from csrank.constants import RANKNET_DC
from csrank.constants import RANKSVM_DC
from csrank.discretechoice import CmpNetDiscreteChoiceFunction
from csrank.discretechoice import FATEDiscreteChoiceFunction
from csrank.discretechoice import FETADiscreteChoiceFunction
from csrank.discretechoice import GeneralizedNestedLogitModel
from csrank.discretechoice import MixedLogitModel
from csrank.discretechoice import MultinomialLogitModel
from csrank.discretechoice import NestedLogitModel
from csrank.discretechoice import PairedCombinatorialLogit
from csrank.discretechoice import PairwiseSVMDiscreteChoiceFunction
from csrank.discretechoice import RankNetDiscreteChoiceFunction
from csrank.metrics_np import categorical_accuracy_np
from csrank.metrics_np import subset_01_loss
from csrank.metrics_np import topk_categorical_accuracy_np
from csrank.util import convert_to_label_encoding
from csrank.util import metrics_on_predictions

metrics = {
    "CategoricalAccuracy": categorical_accuracy_np,
    "CategoricalTopK2": topk_categorical_accuracy_np(k=2),
}


def get_vals(values=[1.0, 1.0]):
    return dict(zip(metrics.keys(), values))


skorch_common_args = {
    "max_epochs": 100,
    "optimizer": optim.SGD,
    "optimizer__lr": 1e-3,
    "optimizer__momentum": 0.9,
    "optimizer__nesterov": True,
    # We evaluate the estimators in-sample. These tests are just small
    # sanity checks, so overfitting is okay here.
    "train_split": None,
}

discrete_choice_functions = {
    CMPNET_DC: (
        CmpNetDiscreteChoiceFunction,
        {"n_hidden": 2, "n_units": 8, **skorch_common_args},
        get_vals([1.0, 1.0]),
    ),
    FATE_DC: (
        FATEDiscreteChoiceFunction,
        {
            "n_hidden_joint_layers": 1,
            "n_hidden_set_layers": 1,
            "n_hidden_joint_units": 5,
            "n_hidden_set_units": 5,
            **skorch_common_args,
        },
        get_vals([1.0, 1.0]),
    ),
    FETA_DC: (
        FETADiscreteChoiceFunction,
        {
            "n_hidden": 1,
            "n_units": 8,
            "add_zeroth_order_model": False,
            **skorch_common_args,
        },
        get_vals([1.0, 1.0]),
    ),
    FETA_DC
    + "zeroth_order_model": (
        FETADiscreteChoiceFunction,
        {
            "n_hidden": 1,
            "n_units": 8,
            "add_zeroth_order_model": True,
            **skorch_common_args,
        },
        get_vals([1.0, 1.0]),
    ),
    MNL: (MultinomialLogitModel, {}, get_vals([0.998, 1.0])),
    NLM: (NestedLogitModel, {}, get_vals()),
    PCL: (PairedCombinatorialLogit, {}, get_vals()),
    GEV: (GeneralizedNestedLogitModel, {}, get_vals()),
    MLM: (MixedLogitModel, {}, get_vals()),
    RANKNET_DC: (
        RankNetDiscreteChoiceFunction,
        {"n_hidden": 2, "n_units": 8, **skorch_common_args},
        get_vals([1.0, 1.0]),
    ),
    RANKSVM_DC: (PairwiseSVMDiscreteChoiceFunction, {}, get_vals([0.982, 0.982])),
}


@pytest.fixture(scope="module")
def trivial_discrete_choice_problem():
    random_state = np.random.RandomState(42)
    # pytorch uses 32 bit floats by default. That should be precise enough and
    # makes it easier to use pytorch and non-pytorch estimators interchangeably.
    x = random_state.randn(500, 5, 2).astype(np.float32)
    w = random_state.rand(2)
    y_true = np.argmax(np.dot(x, w), axis=1)
    y_true = convert_to_label_encoding(y_true, 5)
    return x, y_true


@pytest.mark.parametrize("name", list(discrete_choice_functions.keys()))
def test_discrete_choice_function_fixed(trivial_discrete_choice_problem, name):
    np.random.seed(123)
    # There are some caveats with pytorch reproducibility. See the comment on
    # the corresponding line of `test_choice_functions.py` for details.
    torch.manual_seed(123)
    torch.use_deterministic_algorithms(True)
    x, y = trivial_discrete_choice_problem
    choice_function = discrete_choice_functions[name][0]
    params, accuracies = (
        discrete_choice_functions[name][1],
        discrete_choice_functions[name][2],
    )
    learner = choice_function(**params)
    if name in [MNL, NLM, GEV, PCL, MLM]:
        learner.fit(
            x,
            y,
            vi_params={
                "n": 100,
                "method": "advi",
                "callbacks": [CheckParametersConvergence()],
            },
        )
    else:
        learner.fit(x, y)
    s_pred = learner.predict_scores(x)
    y_pred = learner.predict_for_scores(s_pred)
    y_pred_2 = learner.predict(x)
    rtol = 1e-2
    atol = 5e-2
    assert np.isclose(
        0.0, subset_01_loss(y_pred, y_pred_2), rtol=rtol, atol=atol, equal_nan=False
    )
    for key, value in accuracies.items():
        metric = metrics[key]
        if metric in metrics_on_predictions:
            pred_loss = metric(y, y_pred)
        else:
            pred_loss = metric(y, s_pred)
        assert np.isclose(value, pred_loss, rtol=rtol, atol=atol, equal_nan=False)



from abc import ABCMeta

from csrank.constants import OBJECT_RANKING
from csrank.learner import SkorchInstanceEstimator
from csrank.numpy_util import scores_to_rankings
from csrank.rank_losses import HingedRankLoss

__all__ = ["ObjectRanker", "SkorchObjectRanker"]


class ObjectRanker(metaclass=ABCMeta):
    @property
    def learning_problem(self):
        return OBJECT_RANKING

    def predict_for_scores(self, scores, **kwargs):
        """
        The permutation vector :math:`\\pi` represents the ranking amongst the objects in :math:`Q`, such that
        :math:`\\pi(k)` is the position of the :math:`k`-th object :math:`x_k`, and :math:`\\pi^{-1}(k)` is the index
        of the object on position :math:`k`. Predict rankings for the scores for a given collection of sets of
        objects (query sets).

        Parameters
        ----------
        scores : dict or numpy array
            Dictionary with a mapping from query set size to numpy arrays or a single numpy array of size containing
            scores of each object of size: (n_instances, n_objects)

        Returns
        -------
        Y : dict or numpy array
            Dictionary with a mapping from objects size to numpy arrays or a single numpy array containing
            predicted rankings of size: (n_samples, n_objects)
        """

        if isinstance(scores, dict):
            result = dict()
            for n, score in scores.items():
                rankings = scores_to_rankings(score)
                result[n] = rankings
        else:
            result = scores_to_rankings(scores)
        return result


class SkorchObjectRanker(ObjectRanker, SkorchInstanceEstimator):
    """Base estimator for torch-based ranking.

    This makes it very simple to derive new estimators with any given scoring
    module. Refer to skorch's documentation for supported parameters. For
    example the optimizer or the optimizer's learning rate could be overridden.

    Parameters
    ----------
    module : torch module (class)
        This is the scoring module. It should be an uninstantiated
        ``torch.nn.Module`` class that expects the number of features per
        object as its only parameter on initialization.

    criterion : torch criterion (class)
        The criterion that is used to evaluate and optimize the module.

    **kwargs : skorch NeuralNet arguments
        All keyword arguments are passed to the constructor of
        ``skorch.NeuralNet``. See the documentation of that class for more
        details.
    """

    def __init__(self, module, criterion=HingedRankLoss, **kwargs):
        super().__init__(module=module, criterion=criterion, **kwargs)



import numpy as np

__all__ = [
    "sub_sampling_rankings",
]


def sub_sampling_rankings(Xt, Yt, n_objects=5):
    bucket_size = int(Xt.shape[1] / n_objects)
    X_train = []
    Y_train = []
    for i in range(bucket_size):
        X = np.copy(Xt)
        Y = np.copy(Yt)
        rs = np.random.RandomState(42 + i)
        idx = rs.randint(bucket_size, size=(len(X), n_objects))
        idx += np.arange(start=0, stop=X.shape[1], step=bucket_size)[:n_objects]
        X = X[np.arange(X.shape[0])[:, None], idx]
        Y = Y[np.arange(X.shape[0])[:, None], idx]
        tmp_sort = Y.argsort(axis=-1)
        Y = np.empty_like(Y)
        Y[np.arange(len(X))[:, None], tmp_sort] = np.arange(n_objects)
        if len(X_train) == 0:
            X_train = X
            Y_train = Y
        else:
            Y_train = np.concatenate([Y_train, Y], axis=0)
            X_train = np.concatenate([X_train, X], axis=0)
    return X_train, Y_train



"""High level implementation of scoring modules.

These are pytorch modules that take a list of instances and score each object
within each instance, taking its context into account. Only the high-level
assembly is done in this module.

One can easily derive ranking and choice estimators from a scorer, see
the implemented pytorch estimators such as ``FATEDiscreteObjectChooser`` for
examples.
"""

from .cmpnet import CmpNetScoring
from .fate import FATEScoring
from .feta import FETAScoring
from .ranknet import RankNetScoring

__all__ = ["CmpNetScoring", "FATEScoring", "FETAScoring", "RankNetScoring"]



"""Modules that reduce instances to some kind of feature representation.

The modules listed here should take a 2 or higher dimensional input and reduce
the second-to-last dimension. They can take interaction of "elements" (feature
vectors) into account.

Inputs of shape :math:`(N, *, O, H_i)` are transformed to outputs of shape
:math:`(N, *, H_o)`. In this case :math:`N` is the batch size, :math:`*`
denotes arbitrary additional dimension (which are preserved), :math:`O` refers
to the number of objects per instance (the reduced dimension) and :math:`H_i`
and :math:`H_o` refer to the number of input and output features respectively.
"""

import torch
import torch.nn as nn


class DeepSet(nn.Module):
    """Aggregate object-level embeddings with a mean reduction.

    This module evaluates each object individually (using a object level
    embedding) and then aggregates the embeddings with a mean reduction.

    Parameters
    ----------
    n_features : int
        The number of features per object.

    embedding_size : int
        The target embedding size.

    embedding_module : torch module
        An uninitialized torch module that expects two parameters: the input
        and the output size. It should then act similar to ``nn.Linear``, i.e.
        transform only the last dimension of the input. Defaults to a simple
        linear module.
    """

    def __init__(
        self,
        n_features: int,
        embedding_size: int,
        embedding_module: nn.Module = nn.Linear,
    ):
        super().__init__()
        self.embedding_module = embedding_module(n_features, embedding_size)

    def forward(self, instances):
        """Forward inputs through the network.

        Parameters
        ----------
        instances : tensor
            The input tensor of shape (N, *, O, F), where F is the number of
            features and O is the number of objects.

        Returns
        -------
        tensor
            A tensor of shape (N, *, E), where E ist the embedding size.
        """
        embedded_objects = self.embedding_module(instances)
        return torch.mean(embedded_objects, dim=1)


class MeanAggregatedUtility(nn.Module):
    """Map instances to individual utilities by a first order utility function and aggregation.

    This is an approximation of a generalized utility function as described in
    section 3.1 of [1]_. This is the "Then Aggregate" part of FETA. Only
    aggregation of first-order utility functions is supported.

    Parameters
    ----------
    first_order_utility : torch module (instantiated)
        A function that maps instances to a list of utility matrices. The
        resulting matrix will be interpreted as the utility each object has,
        given the context of the other object.

        The function should expect input in the shape of [N, n_features] and
        return output in the shape of [N]. That is, it should map a list of
        concatenated object features to the pairwise utilities of these
        objects.
    exclude_self_comparison : bool
        Whether or not to exclude the evaluations of the objects with
        themselves (i.e. the diagonal values) when aggregating.

    Returns
    -------
    A list of instances with the aggregated object utilities.

    References
    ----------
    .. [1] Pfannschmidt, K., Gupta, P., & Hüllermeier, E. (2019). Learning
    choice functions: Concepts and architectures. arXiv preprint
    arXiv:1901.10860.
    """

    def __init__(self, first_order_utility, exclude_self_comparison):
        super().__init__()
        self.first_order_utility = first_order_utility
        self.exclude_self_comparison = exclude_self_comparison

    def forward(self, instances, **kwargs):
        """Aggregate zeroth and first order utility, returning a 1d tensor of evaluations.

        Parameters
        ----------
        instances : tensor
            The input tensor of shape (N, *, O, F), where F is the number of
            features and O is the number of objects.

        Returns
        -------
        tensor
            The aggregated utility of each object: A tensor of shape (N, *, O).
        """
        utility_matrix = self.first_order_utility(instances)
        n_objects = utility_matrix.shape[-1]
        n_instances = utility_matrix.shape[0]
        if self.exclude_self_comparison:
            # One diagonal unit matrix for each instance
            diagonal_mask = (
                torch.eye(n_objects, dtype=torch.uint8)
                .view(1, n_objects, n_objects)
                .repeat(n_instances, 1, 1)
            )
            # Fill the diagonal of each instance with 0 (to ignore it while
            # summing).
            utility_matrix[diagonal_mask] = 0
            # The diagonal values do not count when computing the mean.
            n_objects -= 1
        # Compute the mean utility of each object, possibly ignoring the
        # diagonal values.
        context_utility = utility_matrix.sum(dim=-1) / n_objects
        return context_utility



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['csrank',
 'csrank.choicefunction',
 'csrank.core',
 'csrank.dataset_reader',
 'csrank.dataset_reader.choicefunctions',
 'csrank.dataset_reader.discretechoice',
 'csrank.dataset_reader.dyadranking',
 'csrank.dataset_reader.labelranking',
 'csrank.dataset_reader.objectranking',
 'csrank.deprecated',
 'csrank.discretechoice',
 'csrank.dyadranking',
 'csrank.modules',
 'csrank.modules.scoring',
 'csrank.objectranking',
 'csrank.tests']

package_data = \
{'': ['*'], 'csrank.deprecated': ['hotel_dataset/*', 'university_dataset/*']}

install_requires = \
['arviz>=0.10,<0.11.1',
 'numpy>=1.12.1,<2.0.0',
 'pymc3>=3.8,<4.0',
 'scikit-learn>=0.23,<0.24',
 'scipy>=1.2,<2.0',
 'skorch>=0.9,<0.11',
 'torch>=1.8.0,<2.0.0']

extras_require = \
{'data': ['psycopg2-binary>=2.7,<3.0',
          'pandas>=1.1.1,<2.0.0',
          'h5py>=3.0,<4.0',
          'joblib>=0.16.0,<1.0'],
 'docs': ['Sphinx>=3.2.1,<4.0.0',
          'sphinx_rtd_theme>=0.5.0,<0.6.0',
          'sphinxcontrib-bibtex>=1.0.0,<2.0.0',
          'nbsphinx>=0.7.1,<0.8.0',
          'IPython>=7.18.1,<8.0.0'],
 'hypervolume': ['pygmo>=2.7,<3.0']}

setup_kwargs = {
    'name': 'csrank',
    'version': '2.0.0rc2',
    'description': 'Context-sensitive ranking and choice',
    'long_description': '|Coverage| |Binder|\n\n****\nNOTE\n****\n\nThis library has recently been migrated from tensorflow to PyTorch. The 2.0\nversion marks a breaking change. Some of the previous functionality is now\nunavailable and some classes behave differently. You can use the latest 1.x\nrelease if you are looking for the tensorflow based estimators.\n\n*******\nCS-Rank\n*******\n\nCS-Rank is a Python package for context-sensitive ranking and choice\nalgorithms.\n\nWe implement the following new object ranking/choice architectures:\n\n* FATE (First aggregate then evaluate)\n* FETA (First evaluate then aggregate)\n\nIn addition, we also implement these algorithms for choice functions:\n\n* RankNetChoiceFunction\n* GeneralizedLinearModel\n* PairwiseSVMChoiceFunction\n\nThese are the state-of-the-art approaches implemented for the discrete choice\nsetting:\n\n* GeneralizedNestedLogitModel\n* MixedLogitModel\n* NestedLogitModel\n* PairedCombinatorialLogit\n* RankNetDiscreteChoiceFunction\n* PairwiseSVMDiscreteChoiceFunction\n\n\nGetting started\n===============\nAs a simple "Hello World!"-example we will try to learn the Pareto problem:\n\n.. code-block:: python\n\n   import csrank as cs\n   from csrank import ChoiceDatasetGenerator\n   gen = ChoiceDatasetGenerator(dataset_type=\'pareto\',\n                                   n_objects=30,\n                                   n_features=2)\n   X_train, Y_train, X_test, Y_test = gen.get_single_train_test_split()\n\nAll our learning algorithms are implemented using the scikit-learn estimator\nAPI. Fitting our FATENet architecture is as simple as calling the ``fit``\nmethod:\n\n.. code-block:: python\n\n   fate = cs.FATEChoiceFunction()\n   fate.fit(X_train, Y_train)\n\nPredictions can then be obtained using:\n\n.. code-block:: python\n\n   fate.predict(X_test)\n\n\nInstallation\n------------\nThe latest release version of CS-Rank can be installed from Github as follows::\n\n   pip install git+https://github.com/kiudee/cs-ranking.git\n\nAnother option is to clone the repository and install CS-Rank using::\n\n   python setup.py install\n\n\nDependencies\n------------\nCS-Rank depends on PyTorch, skorch, NumPy, SciPy, matplotlib, scikit-learn,\njoblib and tqdm. For data processing and generation you will\nalso need PyGMO, H5Py and pandas.\n\nCiting CS-Rank\n----------------\nYou can cite our `arXiv papers`_::\n\n\n\n  @article{csrank2019,\n    author    = {Karlson Pfannschmidt and\n                 Pritha Gupta and\n                 Eyke H{\\"{u}}llermeier},\n    title     = {Learning Choice Functions: Concepts and Architectures },\n    journal   = {CoRR},\n    volume    = {abs/1901.10860},\n    year      = {2019}\n  }\n\n  @article{csrank2018,\n    author    = {Karlson Pfannschmidt and\n                 Pritha Gupta and\n                 Eyke H{\\"{u}}llermeier},\n    title     = {Deep architectures for learning context-dependent ranking functions},\n    journal   = {CoRR},\n    volume    = {abs/1803.05796},\n    year      = {2018}\n  }\n\nLicense\n--------\n`Apache License, Version 2.0 <https://github.com/kiudee/cs-ranking/blob/master/LICENSE>`_\n\n.. |Binder| image:: https://mybinder.org/badge_logo.svg\n   :target: https://mybinder.org/v2/gh/kiudee/cs-ranking/master?filepath=docs%2Fnotebooks\n\n.. |Coverage| image:: https://codecov.io/gh/kiudee/cs-ranking/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/kiudee/cs-ranking\n\n..\n  |Build Status| image:: https://img.shields.io/github/workflow/status/kiudee/cs-ranking/tests\n  :target: https://github.com/kiudee/cs-ranking/actions\n  :alt: GitHub Workflow Status\n\n\n.. _interactive notebooks: https://mybinder.org/v2/gh/kiudee/cs-ranking/master?filepath=docs%2Fnotebooks\n.. _arXiv papers: https://arxiv.org/search/cs?searchtype=author&query=Pfannschmidt%2C+K\n',
    'author': 'Karlson Pfannschmidt',
    'author_email': 'kiudee@mail.upb.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kiudee/cs-ranking',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import numpy as np

from csrank.dataset_reader import ChoiceDatasetGenerator


def test_pareto_problem_generation():
    """A simple sanity check for Pareto problem generation."""
    gen = ChoiceDatasetGenerator(
        dataset_type="pareto",
        random_state=42,
        n_train_instances=11,
        n_test_instances=1,
        n_objects=3,
        n_features=2,
    )
    X_train, Y_train, X_test, Y_test = gen.get_single_train_test_split()
    assert X_train.shape == (11, 3, 2)
    assert Y_train.shape == (11, 3)
    assert X_test.shape == (1, 3, 2)

    def is_binary_array(a):
        return np.logical_or(a == 0, a == 1).all()

    assert is_binary_array(Y_train)
    assert is_binary_array(Y_test)



import logging

import numpy as np
from scipy.spatial.distance import pdist
from scipy.spatial.distance import squareform
from sklearn.datasets import make_blobs
from sklearn.datasets import make_regression
from sklearn.gaussian_process.kernels import Matern
from sklearn.utils import check_random_state

from csrank.constants import DISCRETE_CHOICE
from csrank.util import convert_to_label_encoding
from ..synthetic_dataset_generator import SyntheticDatasetGenerator
from ..util import create_pairwise_prob_matrix


logger = logging.getLogger(__name__)


class DiscreteChoiceDatasetGenerator(SyntheticDatasetGenerator):
    def __init__(self, dataset_type="medoid", **kwargs):
        super(DiscreteChoiceDatasetGenerator, self).__init__(
            learning_problem=DISCRETE_CHOICE, **kwargs
        )
        dataset_function_options = {
            "linear": self.make_linear_transitive,
            "medoid": self.make_intransitive_medoids,
            "nearest_neighbour_medoid": self.nearest_neighbour,
            "hypervolume": self.make_hv_dataset,
            "gp_transitive": self.make_gp_transitive,
            "gp_non_transitive": self.make_gp_non_transitive,
        }
        if dataset_type not in dataset_function_options:
            raise ValueError(
                f"dataset_type must be one of {set(dataset_function_options.keys())}"
            )
        logger.info("dataset type {}".format(dataset_type))
        self.dataset_function = dataset_function_options[dataset_type]

    def make_linear_transitive(
        self,
        n_instances=1000,
        n_objects=5,
        noise=0.0,
        n_features=100,
        n_informative=10,
        seed=42,
        **kwd,
    ):
        random_state = check_random_state(seed=seed)
        X, y, coeff = make_regression(
            n_samples=n_instances * n_objects,
            n_features=n_features,
            n_informative=n_informative,
            coef=True,
            noise=noise,
            random_state=random_state,
        )
        X = X.reshape(n_instances, n_objects, n_features)
        y = y.reshape(n_instances, n_objects)
        Y = y.argmax(axis=1)
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y

    def make_intransitive_medoids(
        self, n_instances=100, n_objects=5, n_features=100, seed=42, **kwd
    ):
        random_state = check_random_state(seed=seed)
        X = random_state.uniform(size=(n_instances, n_objects, n_features))
        Y = np.empty(n_instances)
        for i in range(n_instances):
            D = squareform(pdist(X[i], metric="euclidean"))
            sum_dist = D.mean(axis=0)
            medoid = np.argmin(sum_dist)
            Y[i] = medoid
        X = np.array(X)
        Y = np.array(Y)
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y

    def nearest_neighbour(
        self, n_instances=100, n_objects=5, n_features=100, seed=42, **kwd
    ):
        random_state = check_random_state(seed=seed)
        X = random_state.uniform(size=(n_instances, n_objects, n_features))
        Y = np.empty(n_instances)
        for i in range(n_instances):
            D = squareform(pdist(X[i], metric="euclidean"))
            sum_dist = D.mean(axis=0)
            medoid = np.argmin(sum_dist)
            Y[i] = np.argsort(D[medoid])[1]
        X = np.array(X)
        Y = np.array(Y)
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y

    def make_hv_dataset(
        self,
        n_instances=1000,
        n_objects=5,
        n_features=5,
        seed=42,
        cluster_spread=1.0,
        **kwd,
    ):
        try:
            from pygmo import hypervolume
        except ImportError:
            from csrank.util import MissingExtraError

            raise MissingExtraError("pygmo", "data")

        def sample_unit_ball(n_f=2, rng=None, radius=1.0):
            rng = check_random_state(rng)
            X = rng.randn(1, n_f)
            u = rng.uniform(size=1)[:, None]
            X /= np.linalg.norm(X, axis=1, ord=2)[:, None]
            X *= radius * u
            return X[0]

        random_state = check_random_state(seed=seed)
        X = random_state.rand(n_instances, n_objects, n_features)
        # Normalize to unit circle and fold to lower quadrant
        X = -np.abs(X / np.sqrt(np.power(X, 2).sum(axis=2))[..., None])
        Y = np.empty(n_instances, dtype=int)
        for i in range(n_instances):
            center = sample_unit_ball(n_f=n_features, rng=i, radius=cluster_spread)
            X[i] = X[i] + center
            hv = hypervolume(X[i])
            cont = hv.contributions(center)
            Y[i] = np.argmax(cont)
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y

    def make_gp_transitive(
        self,
        n_instances=1000,
        n_objects=5,
        noise=0.0,
        n_features=100,
        kernel_params=None,
        seed=42,
        **kwd,
    ):
        """Creates a nonlinear object ranking problem by sampling from a
        Gaussian process as the latent utility function.
        Note that this function needs to compute a kernel matrix of size
        (n_instances * n_objects) ** 2, which could allocate a large chunk of the
        memory."""
        random_state = check_random_state(seed=seed)

        if kernel_params is None:
            kernel_params = dict()
        n_total = n_instances * n_objects
        X = random_state.rand(n_total, n_features)
        L = np.linalg.cholesky(Matern(**kernel_params)(X))
        f = L.dot(random_state.randn(n_total)) + random_state.normal(
            scale=noise, size=n_total
        )
        X = X.reshape(n_instances, n_objects, n_features)
        f = f.reshape(n_instances, n_objects)
        Y = f.argmax(axis=1)
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y

    def make_gp_non_transitive(
        self,
        n_instances=1000,
        n_objects=5,
        n_features=100,
        center_box=(-10.0, 10.0),
        cluster_std=2.0,
        seed=42,
        **kwd,
    ):
        n_samples = n_instances * n_objects
        random_state = check_random_state(seed=seed)
        x, y = make_blobs(
            n_samples=n_samples,
            centers=n_objects,
            n_features=n_features,
            cluster_std=cluster_std,
            center_box=center_box,
            random_state=random_state,
            shuffle=True,
        )
        y = np.array([y])
        samples = np.append(x, y.T, axis=1)
        samples = samples[samples[:, n_features].argsort()]
        pairwise_prob = create_pairwise_prob_matrix(n_objects)
        X = []
        Y = []
        for inst in range(n_instances):
            feature = np.array(
                [samples[inst + i * n_instances, 0:-1] for i in range(n_objects)]
            )
            matrix = np.random.binomial(1, pairwise_prob)
            choice = np.argmax(np.sum(matrix, axis=1))
            X.append(feature)
            Y.append(choice)
        X = np.array(X)
        Y = np.array(Y)
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y



from .choice_data_generator import ChoiceDatasetGenerator
from .expedia_choice_dataset_reader import ExpediaChoiceDatasetReader
from .letor_ranking_choice_dataset_reader import LetorRankingChoiceDatasetReader
from .mnist_choice_dataset_reader import MNISTChoiceDatasetReader

__all__ = [
    "ChoiceDatasetGenerator",
    "ExpediaChoiceDatasetReader",
    "LetorRankingChoiceDatasetReader",
    "MNISTChoiceDatasetReader",
]



from datetime import datetime
from datetime import timedelta
import inspect
import logging
import os
from pathlib import Path
import re
import sys

import numpy as np
from sklearn.preprocessing import LabelBinarizer

from csrank.metrics_np import f1_measure
from csrank.metrics_np import hamming
from csrank.metrics_np import instance_informedness
from csrank.metrics_np import precision
from csrank.metrics_np import recall
from csrank.metrics_np import subset_01_loss
from csrank.metrics_np import zero_one_accuracy_np

__all__ = [
    "create_dir_recursively",
    "duration_till_now",
    "print_dictionary",
    "rename_file_if_exist",
    "seconds_to_time",
    "time_from_now",
    "get_duration_seconds",
    "setup_logging",
    "progress_bar",
]


metrics_on_predictions = [
    f1_measure,
    precision,
    recall,
    subset_01_loss,
    hamming,
    instance_informedness,
    zero_one_accuracy_np,
]


class MissingExtraError(ImportError):
    """Indicates an ImportError that can be fixed by installing an extra"""

    def __init__(self, package, extra):
        super(MissingExtraError, self).__init__(
            f"Could not import the optional dependency {package}. "
            f'Please install it or specify the "{extra}" extra when installing this package.'
        )


def progress_bar(count, total, status=""):
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    bar = "=" * filled_len + "-" * (bar_len - filled_len)

    sys.stdout.write("[%s] %s/%s ...%s\r" % (bar, count, total, status))
    sys.stdout.flush()


def print_dictionary(dictionary):
    output = "\n"
    for key, value in dictionary.items():
        output = output + str(key) + " => " + str(value) + "\n"
    return output


def create_dir_recursively(path, is_file_path=False):
    if is_file_path:
        path = os.path.dirname(path)
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def rename_file_if_exist(file_path):
    my_file = Path(file_path)
    try:
        extension = "." + file_path.split(".")[1]
    except IndexError:
        extension = ""
    path = file_path.split(".")[0]
    i = 1
    while my_file.is_file():
        file_path = path + str(i) + extension
        my_file = Path(file_path)
        i += 1
    return file_path


def get_duration_seconds(duration):
    time = int(re.findall(r"\d+", duration)[0])
    d = duration.split(str(time))[1].upper()
    options = {"D": 24 * 60 * 60, "H": 60 * 60, "M": 60}
    return options[d] * time


def duration_till_now(start):
    return (datetime.now() - start).total_seconds()


def time_from_now(target_time_sec):
    base_datetime = datetime.now()
    delta = timedelta(seconds=target_time_sec)
    target_date = base_datetime + delta
    return target_date.strftime("%Y-%m-%d %H:%M:%S")


def seconds_to_time(target_time_sec):
    return str(timedelta(seconds=target_time_sec))


def convert_to_loss(loss_function):
    def loss(y_true, y_pred):
        return 1.0 - loss_function(y_true, y_pred)

    return loss


def setup_logging(log_path=None, level=logging.DEBUG):
    """Function setup as many logging for the experiments"""
    if log_path is None:
        dirname = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        dirname = os.path.dirname(dirname)
        log_path = os.path.join(dirname, "experiments", "logs", "logs.log")
        create_dir_recursively(log_path, True)
    logging.basicConfig(
        filename=log_path,
        level=level,
        format="%(asctime)s %(name)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logger = logging.getLogger("SetupLogger")
    logger.info("log file path: {}".format(log_path))


def convert_to_label_encoding(Y, n_objects):
    lb = LabelBinarizer().fit(np.arange(n_objects))
    return lb.transform(Y)



import logging

import numpy as np

from csrank.learner import Learner
import csrank.theano_util as ttu
from csrank.util import print_dictionary
from .discrete_choice import DiscreteObjectChooser
from .likelihoods import create_weight_dictionary
from .likelihoods import fit_pymc3_model
from .likelihoods import likelihood_dict
from .likelihoods import LogLikelihood

try:
    import pymc3 as pm
    from pymc3.variational.callbacks import CheckParametersConvergence
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")

logger = logging.getLogger(__name__)


class MultinomialLogitModel(DiscreteObjectChooser, Learner):
    def __init__(self, loss_function="", regularization="l2", **kwargs):
        """
        Create an instance of the Multinomial Logit model for learning the discrete choice function. The utility
        score for each object in query set :math:`Q` is defined as :math:`U(x) = w \\cdot x`, where :math:`w` is
        the weight vector. The probability of choosing an object :math:`x_i` is defined by taking softmax over the
        utility scores of the objects:

        .. math::

            P(x_i \\lvert Q) = \\frac{exp(U(x_i))}{\\sum_{x_j \\in Q} exp(U(x_j))}

        The discrete choice for the given query set :math:`Q` is defined as:

        .. math::

            dc(Q) := \\operatorname{argmax}_{x_i \\in Q }  \\; P(x_i \\lvert Q)

        Parameters
        ----------
        loss_function : string , {‘categorical_crossentropy’, ‘binary_crossentropy’, ’categorical_hinge’}
            Loss function to be used for the discrete choice decision from the query set
        regularization : string, {‘l1’, ‘l2’}, string
           Regularizer function (L1 or L2) applied to the `kernel` weights matrix
        random_state : int or object
            Numpy random state
        **kwargs
            Keyword arguments for the algorithms

        References
        ----------
            [1] Kenneth E Train. „Discrete choice methods with simulation“. In: Cambridge university press, 2009. Chap Logit, pp. 41–86.

            [2] Kenneth Train. Qualitative choice analysis. Cambridge, MA: MIT Press, 1986
        """
        self.loss_function = loss_function
        known_regularization_functions = {"l1", "l2"}
        if regularization not in known_regularization_functions:
            raise ValueError(
                f"Regularization function {regularization} is unknown. Must be one of {known_regularization_functions}"
            )
        self.regularization = regularization

    @property
    def model_configuration(self):
        """
            Constructs the dictionary containing the priors for the weight vectors for the model according to the
            regularization function. The parameters are:
                * **weights** : Weights to evaluates the utility of the objects

            For ``l1`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{b}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Laplace}(\\text{mu}=\\text{mu}_w, \\text{b}=\\text{b}_w)

            For ``l2`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{sd}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Normal}(\\text{mu}=\\text{mu}_w, \\text{sd}=\\text{sd}_w)
        """
        if not hasattr(self, "config_"):
            if self.regularization == "l2":
                weight = pm.Normal
                prior = "sd"
            elif self.regularization == "l1":
                weight = pm.Laplace
                prior = "b"
            self.config_ = {
                "weights": [
                    weight,
                    {
                        "mu": (pm.Normal, {"mu": 0, "sd": 5}),
                        prior: (pm.HalfCauchy, {"beta": 1}),
                    },
                ]
            }
            logger.info(
                "Creating model with config {}".format(print_dictionary(self.config_))
            )
        return self.config_

    def construct_model(self, X, Y):
        """
        Constructs the multinomial logit model which evaluated the utility score as :math:`U(x) = w \\cdot x`, where
        :math:`w` is the weight vector. The probability of choosing the object :math:`x_i` from the query set
        :math:`Q = \\{x_1, \\ldots ,x_n\\}` is:

        .. math::

            P_i = P(x_i \\lvert Q) = \\frac{exp(U(x_i))}{\\sum_{x_j \\in Q} exp(U(x_j))}

        Parameters
        ----------
        X : numpy array
            (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array
            (n_instances, n_objects)
            Preferences in the form of discrete choices for given objects

        Returns
        -------
         model : pymc3 Model :class:`pm.Model`
        """
        self.trace_ = None
        self.trace_vi_ = None
        logger.info(
            "Creating model_args config {}".format(
                print_dictionary(self.model_configuration)
            )
        )
        self.loss_function_ = likelihood_dict.get(self.loss_function, None)
        with pm.Model() as self.model:
            self.Xt_ = theano.shared(X)
            self.Yt_ = theano.shared(Y)
            shapes = {"weights": self.n_object_features_fit_}
            # shapes = {'weights': (self.n_object_features_fit_, 3)}
            weights_dict = create_weight_dictionary(self.model_configuration, shapes)
            intercept = pm.Normal("intercept", mu=0, sd=10)
            utility = tt.dot(self.Xt_, weights_dict["weights"]) + intercept
            self.p_ = ttu.softmax(utility, axis=1)

            LogLikelihood(
                "yl", loss_func=self.loss_function_, p=self.p_, observed=self.Yt_
            )
        logger.info("Model construction completed")

    def fit(
        self,
        X,
        Y,
        sampler="variational",
        tune=500,
        draws=500,
        vi_params={
            "n": 20000,
            "method": "advi",
            "callbacks": [CheckParametersConvergence()],
        },
        **kwargs,
    ):
        """
        Fit a multinomial logit model on the provided set of queries X and choices Y of those objects. The
        provided queries and corresponding preferences are of a fixed size (numpy arrays). For learning this network
        the categorical cross entropy loss function for each object :math:`x_i \\in Q` is defined as:

        .. math::

            C_{i} =  -y(i)\\log(P_i) \\enspace,

        where :math:`y` is ground-truth discrete choice vector of the objects in the given query set :math:`Q`.
        The value :math:`y(i) = 1` if object :math:`x_i` is chosen else :math:`y(i) = 0`.

        Parameters
        ----------
        X : numpy array (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array (n_instances, n_objects)
            Choices for given objects in the query
        sampler : {‘variational’, ‘metropolis’, ‘nuts’}, string
            The sampler used to estimate the posterior mean and mass matrix from the trace

                * **variational** : Run inference methods to estimate posterior mean and diagonal mass matrix
                * **metropolis** : Use the MAP as starting point and Metropolis-Hastings sampler
                * **nuts** : Use the No-U-Turn sampler
        vi_params : dict
            The parameters for the **variational** inference method
        draws : int
            The number of samples to draw. Defaults to 500. The number of tuned samples are discarded by default
        tune : int
            Number of iterations to tune, defaults to 500. Ignored when using 'SMC'. Samplers adjust
            the step sizes, scalings or similar during tuning. Tuning samples will be drawn in addition
            to the number specified in the `draws` argument, and will be discarded unless
            `discard_tuned_samples` is set to False.
        **kwargs :
            Keyword arguments for the fit function of :meth:`pymc3.fit`or :meth:`pymc3.sample`
        """
        self._pre_fit()
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        self.construct_model(X, Y)
        fit_pymc3_model(self, sampler, draws, tune, vi_params, **kwargs)
        return self

    def _predict_scores_fixed(self, X, **kwargs):
        d = dict(pm.summary(self.trace_)["mean"])
        intercept = 0.0
        weights = np.array(
            [d["weights[{}]".format(i)] for i in range(self.n_object_features_fit_)]
        )
        if "intercept" in d:
            intercept = intercept + d["intercept"]
        return np.dot(X, weights) + intercept



"""Modules that transform feature vectors individually.

The modules listed here should take a 1 or higher dimensional input and apply
some mapping to the last dimension.  They do not take interactions of different
feature vectors into account.

Inputs of shape :math:`(N, *, H_i)` are transformed to outputs of shape
:math:`(N, *, H_o)`. In this case :math:`N` is the batch size, :math:`*`
denotes arbitrary additional dimension (which are preserved), :math:`H_i`
refers to the number of input features :math:`H_o` refers to the number of
input and output features respectively.
"""

import torch
import torch.nn as nn

# Refer to Figure 1 of https://arxiv.org/pdf/1901.10860.pdf for an overview of
# the different components that are used for FATE and FETA.


class DeterministicSumming(nn.Module):
    """Transform a tensor into repetitions of its sum.

    Intended for use in tests, not useful for actual learning. The last
    dimension of the input should contain feature vectors. The result will be
    an array of matching shape with the last dimension replaced by repeated
    utility values (i.e. sums).

    Let's use this as a pairwise utility function. As an example, consider
    this pairing. There are two instances with two objects each. All object
    combinations are considered. Objects have two features.

    >>> import torch
    >>> pairs = torch.tensor(
    ...    [[[0.5000, 0.6000, 0.5000, 0.6000],
    ...      [0.5000, 0.6000, 1.5000, 1.6000],
    ...      [1.5000, 1.6000, 0.5000, 0.6000],
    ...      [1.5000, 1.6000, 1.5000, 1.6000]],
    ...     [[2.5000, 2.6000, 2.5000, 2.6000],
    ...      [2.5000, 2.6000, 3.5000, 3.6000],
    ...      [3.5000, 3.6000, 2.5000, 2.6000],
    ...      [3.5000, 3.6000, 3.5000, 3.6000]]])

    We can compute the mock utility of this pairing as follows:

    >>> utility = DeterministicSumming(input_size=2)
    >>> utilities = utility(pairs)
    >>> utilities
    tensor([[[ 2.2000],
             [ 4.2000],
             [ 4.2000],
             [ 6.2000]],
    <BLANKLINE>
            [[10.2000],
             [12.2000],
             [12.2000],
             [14.2000]]])

    Note that for example :math:`2.2 = 0.5 + 0.6 + 0.5 + 0.6`, that is

    >>> utilities[0][0] == pairs[0][0].sum()
    tensor([True])

    Parameters
    ----------
    input_size : int
        The size of the last dimension of the input.

    output_size : int
        The size of the last dimension of the output. Defaults to `1` to make
        it more convenient to use this as a utility.
    """

    def __init__(self, input_size: int, output_size: int = 1):
        super().__init__()
        self.output_size = output_size

    def forward(self, inputs):
        """Forward inputs through the network.

        Parameters
        ----------
        inputs : tensor
            The input tensor of shape (N, *, I), where I is the input size.

        Returns
        -------
        tensor
            A tensor of shape (N, *, O), where O is the output size.
        """
        summed = inputs.sum(dim=-1)
        # repeat in newly created last dimension
        repeated = (
            summed.view(-1, 1)
            .repeat(1, self.output_size)
            .view(summed.shape + (self.output_size,))
        )
        return repeated


class DenseNeuralNetwork(nn.Module):
    """Deep, densely connected neural network.

    All hidden layers have the same number of units.

    Parameters
    ----------
    input_size: int
        The number of units at the input layer.
    output_size: int
        The number of units at the output layer.
    hidden_layers: int
        The number of hidden layers in addition to the input and output layer.
    units_per_hidden: int
        The number of units each hidden layer has.
    activation: torch activation function
        The activation function that should be applied after each layer.
        Defaults to an instance of `nn.SELU`.
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_layers: int,
        units_per_hidden: int,
        activation=None,
    ):
        super().__init__()
        self.input_layer = nn.Linear(input_size, units_per_hidden)
        # ModuleList is necessary to make pytorch aware of these layers and add
        # their parameters to this module's parameter list.
        self.hidden_layers = nn.ModuleList(
            [
                nn.Linear(units_per_hidden, units_per_hidden)
                for _ in range(hidden_layers)
            ]
        )
        self.output_layer = nn.Linear(units_per_hidden, output_size)
        self.activation = activation if activation is not None else nn.SELU()

    def forward(self, x):
        """Forward inputs through the network.

        Parameters
        ----------
        inputs : tensor
            The input tensor of shape (N, *, I), where I is the input size.

        Returns
        -------
        tensor
            A tensor of shape (N, *, O), where O is the output size.
        """
        result = self.activation(self.input_layer(x))
        for layer in self.hidden_layers:
            result = self.activation(layer(result))
        result = self.output_layer(result)
        return result


def _flip_pairs(pairs):
    """Flip the order of concatenated feature vectors.

    Given a tensor of concatenated feature vectors, such as this one

    >>> pairs = torch.tensor([
    ...     [1,  2,  3,  4],
    ...     [5,  6,  7,  8],
    ...     [9, 10, 11, 12],
    ... ])

    this function will return a tensor of concatenated feature vectors where
    the order of the objects is reversed:

    >>> _flip_pairs(pairs)
    tensor([[ 3,  4,  1,  2],
            [ 7,  8,  5,  6],
            [11, 12,  9, 10]])

    Parameters
    ----------
    pairs: tensor of dimension N >= 1
        The tensor of pairs. It is assumed that the last dimension consists of
        concatenated feature vectors of equal length, i.e. that its size is
        even.

    Returns
    -------
    tensor: of dimension N
        The tensor of pairs with the order of the feature vectors reversed.
    """
    previous_shape = pairs.shape
    assert previous_shape[-1] % 2 == 0
    # split concatenated objects
    new_shape = previous_shape[:-1] + (2, previous_shape[-1] // 2)
    split_pairs = pairs.view(*new_shape)

    # reverse the order of the objects
    reverse_split_pairs = split_pairs.flip(-2)

    # concatenate the feature vectors again
    reverse_pairs = reverse_split_pairs.view(pairs.shape)
    return reverse_pairs


class CmpNN(nn.Module):
    r"""A pairwise preference module with weight sharing.

    This module lifts a "core preference module" to a comparative neural
    network with weight sharing. It uses shared weights for the computations
    :math:`U_1(x_1, x_2)` and :math:`U_1(x_2, x_1)`, which should produce a
    more consistent result. The architecture was originally proposed in [1]_.
    That original proposal suggested a single linear layer for the "core
    preference module", but this implementation generalizes the approach to any
    other "zeroth order module".

    The architecture works by evaluating the "core" network on the object
    pairing in both orders (:math:`U_1(x_1, x_2)` and :math:`U_1(x_2, x_1)`)
    and then aggregating the result. The weights of the two evaluations are
    shared, therefore passing the objects in the opposite order will use the
    same weights for the pairwise evaluations and then aggregate those
    evaluations with swapped weights.

    According to the original architecture this network should return two
    outputs: One that indicates the "evidence" that the first input is
    preferable to the second one, and one that indicates the opposite. The two
    outputs would be swapped when the inputs are swapped. For ease of
    integration with other utility modules, we have decided to only return the
    first output. The other output can be accessed by swapping the inputs. That
    is a bit less efficient for inference, but makes it easier to use this as
    an exchangeable component.

    Parameters
    ----------
    pairwise_preference_core: instantiated pytorch module
        This module is used for the pairwise evaluations. The original paper
        [1]_ suggests to use a simple linear layer. This module should expect
        an input of shape :math:`(N, 2 \cdot F)` where :math:`F` is the size of
        a feature vector, and produce an output of shape :math:`(N, C)` where
        :math:`C` is the core output size.
    core_encoding_size: int
        The size of an object encoding that is returned by the pairwise
        preference core.

    References
    ----------
    .. [1] Rigutini, L., Papini, T., Maggini, M., & Scarselli, F. (2011).
    SortNet: Learning to rank by a neural preference function. IEEE
    transactions on neural networks, 22(9), 1368-1380.
    """

    def __init__(self, pairwise_preference_core, core_encoding_size):
        super().__init__()
        # The core of the network. Just one layer in the original paper. See
        # Figure 1 of [1]. Should expect input size 2*n_features and have
        # output size n_hidden.
        self.pairwise_preference_core = pairwise_preference_core
        # Predict the outputs N_<, N_> (depending on the order in which the inputs are passed)
        self.final_layer = nn.Linear(core_encoding_size * 2, 1)
        # self.pairwise_comparison_module = pairwise_comparison_module

    def forward(self, pairs):
        reverse_pairs = _flip_pairs(pairs)
        comparisons = self.pairwise_preference_core(pairs)
        # Both hidden outputs are aggregated in final layer
        reverse_comparisons = self.pairwise_preference_core(reverse_pairs)
        # "x preferable to y", denoted N_> in the paper
        output_one = self.final_layer(torch.cat([comparisons, reverse_comparisons], -1))
        # "y preferable to x", denoted N_< in the paper

        # output_two is equal to output_one of the reversed input (by design).
        # We could use this to build up the pairwise comparison matrix. That
        # would be more efficient, but also a bit more complex. It should only
        # make a difference for inference.

        # output_two = self.final_layer(torch.cat([reverse_comparisons, comparisons], -1))
        # return torch.cat([output_one, output_two], -1)

        # Note that if the inputs are switched then "comparisons" and "reverse
        # comparisons" are switched and therefore also output_one and
        # output_two are switched. That ensures consistency (i.e. that the
        # preference is independent of the order in which the inputs where
        # given).
        return output_one



import os

import numpy as np
from sklearn.impute import SimpleImputer
from sklearn.model_selection import ShuffleSplit
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_random_state

from csrank.constants import LABEL_RANKING
from csrank.numpy_util import ranking_ordering_conversion
from ..dataset_reader import DatasetReader

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")


class SurveyDatasetReader(DatasetReader):
    def __init__(self, random_state=None, **kwargs):
        super(SurveyDatasetReader, self).__init__(
            learning_problem=LABEL_RANKING, dataset_folder="survey_data", **kwargs
        )
        self.train_file = os.path.join(self.dirname, "rawdata_all.dta")
        self.random_state = check_random_state(random_state)
        self.__load_dataset__()

    def __load_dataset__(self):
        df = pd.io.stata.read_stata(self.train_file)
        orderings = []
        features = []
        for row in df.itertuples():
            orderings.append(row[4:8])
            context_feature = [float(i) if i != "." else np.NAN for i in row[13:33]]
            features.append(context_feature)
        X = np.array(features)
        X = SimpleImputer().fit_transform(X)
        X = np.array([np.log(np.array(X[:, i]) + 1) for i in range(len(features[0]))])
        X = np.array(X.T)
        self.X = StandardScaler().fit_transform(X)
        orderings = np.array(orderings) - 1
        self.Y = ranking_ordering_conversion(orderings)
        self.__check_dataset_validity__()

    def get_single_train_test_split(self):
        cv_iter = ShuffleSplit(
            n_splits=1, test_size=0.3, random_state=self.random_state
        )
        (train_idx, test_idx) = list(cv_iter.split(self.X))[0]
        return self.X[train_idx], self.Y[train_idx], self.X[test_idx], self.Y[test_idx]

    def get_dataset_dictionaries(self):
        return self.X, self.Y
